import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.*;

public final class Main {

  public static final int WIDTH = 640;
  public static final int HEIGHT = 512;

  public static final double[] EYE = { 0, 0, 0 };
  public static final double[] LOOK = { 0, 0, -1 };
  public static final double VIRTUAL_SCREEN_WIDTH = 100;
  public static final double VIRTUAL_SCREEN_HEIGHT
      = VIRTUAL_SCREEN_WIDTH * HEIGHT / (double)WIDTH;
  public static final double VIEWER_DISTANCE = VIRTUAL_SCREEN_WIDTH;
  public static final double TINY = 0.0000000001;
  public static final double EPSILON = 0.0000001;

  public static final int SQRT_SAMPLES = 10;
  public static final int SAMPLES = SQRT_SAMPLES * SQRT_SAMPLES;
  public static final double INVERSE_SAMPLES = 1.0 / SAMPLES;
  public static final String IMAGE_TYPE = "jpg";
  public static final String OUTPUT_FILE = "output." + IMAGE_TYPE;
  public static final int NOISE_SIZE = 1 << 8;
  public static final int NOISE_MASK = NOISE_SIZE - 1;
  public static final long NOISE_SEED = 0xDEADBEEFCAFEBABEL;
  public static final int FBM_ITERATIONS = 16;

  public static final int PALETTE_SIZE = 256;

  public static final double SKY_PALETTE_SCALE = 20;
  public static final double SKY_PALETTE_OFFSET = -12;
  public static final int SKY_BLUE = 0x3D66C2;
  public static final int CLOUD_WHITE = 0xFDFEF8;
  public static final double SKY_SCALE = 0.1;
  public static final double[] SKY_ORIGIN = { 0.0, 200.0, 0.0 };

  public static final double GROUND_PLANT_SCALE = 50;
  public static final double GROUND_PLANT_OFFSET = 255;
  public static final double GROUND_DIRT_SCALE = 30;
  public static final double GROUND_DIRT_OFFSET = 128;
  public static final double GROUND_ROCK_SCALE = 40;
  public static final double GROUND_ROCK_OFFSET = 0;

  public static final int GROUND_PLANT = 0x9BAD57;
  public static final int GROUND_DIRT = 0x817A6A;
  public static final int GROUND_ROCK = 0xC6D1D7;

  public static final int RED = 0;
  public static final int GREEN = 1;
  public static final int BLUE = 2;

  public static final int FBM_GAIN = 0;
  public static final int FBM_LACUNARITY = 1;
  public static final int FBM_SCALE = 2;

  public static final double GAMMA = 2.2;
  public static final double INVERSE_GAMMA = 1.0 / GAMMA;

  public static final long SECOND_MILLIS = 1000L;
  public static final long MINUTE_MILLIS = 60 * SECOND_MILLIS;
  public static final long HOUR_MILLIS = 60 * MINUTE_MILLIS;

  private double[][][] pixels = new double[HEIGHT][WIDTH][3];

  private double[][][] noise = new double[NOISE_SIZE][NOISE_SIZE][NOISE_SIZE];

  private double[][] skyPalette = new double[PALETTE_SIZE][3];
  private double[] fbmSkyMaterial = { 0.4, 2.5, 0.005 };

  private double[][] groundPalette = new double[PALETTE_SIZE][3];
  private double[] fbmGroundMaterial = { 0.3, 3, 0.04 };

  private double[] UP = { 0, 1, 0 };
  private double[] DOWN = { 0, -1, 0 };

  private double[][][] SAMPLE_POINTS = new double[263][SAMPLES][2];

  private class Intersection {
    public double[] o = new double[3];
    public double[] d = new double[3];
    public double[] n = new double[3];
    public double[] color = new double[3];
    public double[] material = new double[3];
    public double[] p = new double[3];
    public double t;
    public boolean hit;

    public double[][] temps = new double[16][3];
  }

  private void launch() throws Throwable {

    initializeNoise();
    initializeSkyPalette();
    initializeGroundPalette();
    initializeSamples();

    render();

    saveImage();

  }

  private void render() {

    // Find orthonormal basis for screen coordinates
    double[] u = new double[3];
    double[] v = new double[3];
    double[] w = new double[3];
    subtract(w, EYE, LOOK);
    normalize(w, w);

    if (equals(w, UP)) {
      set(u, 1, 0, 0);
      set(v, 0, 0, -1);
    } else if (equals(w, DOWN)) {
      set(u, 1, 0, 0);
      set(v, 0, 0, 1);
    } else {
      cross(u, UP, w);
      cross(v, w, u);
    }

    // Find center of the screen in space
    double[] c = new double[3];
    scale(c, w, -VIEWER_DISTANCE);
    add(c, c, EYE);

    double[] vx = new double[3];
    double[] vy = new double[3];

    double[] halfWidth = new double[3];
    set(halfWidth, u);
    scale(halfWidth, halfWidth, VIRTUAL_SCREEN_WIDTH / 2.0);

    double[] halfHeight = new double[3];
    set(halfHeight, v);
    scale(halfHeight, halfHeight, VIRTUAL_SCREEN_HEIGHT / 2.0);

    // Move c to lower-left corner of screen
    subtract(c, c, halfWidth);
    subtract(c, c, halfHeight);

    final double CORNER_X_SCALE = VIRTUAL_SCREEN_WIDTH / WIDTH;
    final double CORNER_Y_SCALE = VIRTUAL_SCREEN_HEIGHT / HEIGHT;

    Intersection intersection = new Intersection();
    double[] d = intersection.d;

    int samplesIndex = 0;
    for(int y = 0; y < HEIGHT; y++) {
      for(int x = 0; x < WIDTH; x++) {

        double[] pixel = pixels[HEIGHT - y - 1][x];

        double[][] samples = SAMPLE_POINTS[samplesIndex];
        if (++samplesIndex == SAMPLE_POINTS.length) {
          samplesIndex = 0;
        }
        for(int i = 0; i < SAMPLES; i++) {

          double[] p = samples[i];

          double cx = (x + p[0]) * CORNER_X_SCALE;
          double cy = (y + p[1]) * CORNER_Y_SCALE;

          set(vx, u);
          scale(vx, vx, cx);

          set(vy, v);
          scale(vy, vy, cy);

          add(d, c, vx);
          add(d, d, vy);
          subtract(d, d, EYE);
          normalize(d, d);

          if (equals(d[1], 0)) {
            d[1] = TINY;
          }

          set(intersection.o, EYE);

          intersectSky(intersection);
          if (intersection.hit) {
            add(pixel, pixel, intersection.color);
          }
        }
      }
    }
  }

  private void intersectSky(Intersection intersection) {

    intersection.hit = false;
   
    intersection.t = (SKY_ORIGIN[1] - intersection.o[1]) / intersection.d[1];
    if (intersection.t < EPSILON) {
      return;
    }

    intersection.hit = true;

    intersection.p[0] = intersection.o[0] + intersection.t * intersection.d[0];
    intersection.p[1] = SKY_ORIGIN[1];
    intersection.p[2] = intersection.o[2] + intersection.t * intersection.d[2];

    double x = SKY_SCALE * (intersection.p[0] - SKY_ORIGIN[0]);
    double z = SKY_SCALE * (intersection.p[2] - SKY_ORIGIN[2]);
    
    if (checkerBoardTest(x, z, 50)) {
      set(intersection.color, 0, 0, 0);
    } else {
      set(intersection.color, 1, 1, 1);
    }
    //set(intersection.color, skyPalette[toPaletteIndex(
    //    fractionalBrownianMotion(x, 0, z, fbmSkyMaterial))]);

    set(intersection.n, 0, -1, 0);
  }

  private void zero(double[] v) {
    for(int i = v.length - 1; i >= 0; i--) {
      v[i] = 0;
    }
  }

  private void set(double[] v, double[] a) {
    for(int i = a.length - 1; i >= 0; i--) {
      v[i] = a[i];
    }
  }

  private void set(double[] v, double x, double y, double z) {
    v[0] = x;
    v[1] = y;
    v[2] = z;
  }
  
  private boolean equals(double[] a, double[] b) {
    for(int i = a.length - 1; i >= 0; i--) {
      if (!equals(a[i], b[i])) {
        return false;
      }
    }
    return true;
  }

  private boolean equals(double a, double b) {
    return Math.abs(a - b) < TINY;
  }

  private void negate(double[] d, double[] a) {
    scale(d, a, -1);
  }

  private void scale(double[] d, double[] a, double s) {
    for(int i = a.length - 1; i >= 0; i--) {
      d[i] = s * a[i];
    }
  }

  private void subtract(double[] d, double[] a, double[] b) {
    for(int i = a.length - 1; i >= 0; i--) {
      d[i] = a[i] - b[i];
    }
  }

  private void add(double[] d, double[] a, double[] b) {
    for(int i = a.length - 1; i >= 0; i--) {
      d[i] = a[i] + b[i];
    }
  }

  private void cross(double[] d, double[] a, double[] b) {
    double x = a[1] * b[2] - a[2] * b[1];
    double y = a[2] * b[0] - a[0] * b[2];
    double z = a[0] * b[1] - a[1] * b[0];
    d[0] = x;
    d[1] = y;
    d[2] = z;
  }

  private void dot(double[] d, double[] a, double[] b) {
    for(int i = a.length - 1; i >= 0; i--) {
      d[i] = a[i] * b[i];
    }
  }

  private double inverseMag(double[] v) {
    return 1.0 / mag(v);
  }

  private double mag(double[] v) {
    return Math.sqrt(mag2(v));
  }

  private double mag2(double[] v) {
    double sum = 0;
    for(int i = v.length - 1; i >= 0; i--) {
      sum += v[i] * v[i];
    }
    return sum;
  }

  private void normalize(double[] d, double[] v) {
    scale(d, v, inverseMag(v));
  }

  private void initializeSamples() {
    double scale = 1.0 / SQRT_SAMPLES;
    Random random = new Random(NOISE_SEED);
    for(int i = 0; i < SAMPLE_POINTS.length; i++) {
      for(int y = 0, k = 0; y < SQRT_SAMPLES; y++) {
        for(int x = 0; x < SQRT_SAMPLES; x++, k++) {
          double[] p = SAMPLE_POINTS[i][k];
          if (SQRT_SAMPLES == 1) {
            p[0] = scale * (x + 0.5);
            p[1] = scale * (y + 0.5);
          } else {
            p[0] = scale * (x + random.nextDouble());
            p[1] = scale * (y + random.nextDouble());
          }
        }
      }
    }
  }

  private void initializeGroundPalette() {
    double[] groundPlant = new double[3];
    double[] groundDirt = new double[3];
    double[] groundRock = new double[3];
    convertHexColorToDoubles(groundPlant, GROUND_PLANT);
    convertHexColorToDoubles(groundDirt, GROUND_DIRT);
    convertHexColorToDoubles(groundRock, GROUND_ROCK);
    for(int i = 0; i < PALETTE_SIZE; i++) {
      double ps
          = normalDistribution(i, GROUND_PLANT_SCALE, GROUND_PLANT_OFFSET);
      double ds
          = normalDistribution(i, GROUND_DIRT_SCALE, GROUND_DIRT_OFFSET);     
      double rs
          = normalDistribution(i, GROUND_ROCK_SCALE, GROUND_ROCK_OFFSET);
      for(int j = 0; j < 3; j++) {
        groundPalette[i][j] = bind(ps * groundPlant[j] + ds * groundDirt[j]
            + rs * groundRock[j]);
      }
    }
  }

  private double bind(double x) {
    return bind(x, 0, 1);
  }

  private double bind(double x, double min, double max) {
    if (x < min) {
      x = min;
    } else if (x > max) {
      x = max;
    }
    return x;
  }

  private int bindInt(int x, int min, int max) {
    if (x < min) {
      x = min;
    } else if (x > max) {
      x = max;
    }
    return x;
  }

  private double normalDistribution(double x, double scale, double origin) {
    x = (x - origin) / scale;
    return Math.exp(-0.5 * x * x);
  }

  private void initializeSkyPalette() {
    double[] skyBlue = new double[3];
    double[] cloudWhite = new double[3];
    convertHexColorToDoubles(cloudWhite, CLOUD_WHITE);
    convertHexColorToDoubles(skyBlue, SKY_BLUE);
    for(int i = 0; i < PALETTE_SIZE; i++) {
      double value = 0.5 + Math.atan(SKY_PALETTE_OFFSET + SKY_PALETTE_SCALE * i
          / (double)(PALETTE_SIZE - 1)) / (Math.PI / 2) / 2;
      value = bind(value);
      for(int j = 0; j < 3; j++) {
        skyPalette[i][j] = cloudWhite[j] + (skyBlue[j] - cloudWhite[j]) * value;
      }
    }
  }

  private void convertHexColorToDoubles(double[] output, int input) {
    for(int i = 0; i < 3; i++) {
      int hex = input & 0xFF;
      input >>= 8;
      output[2 - i] = Math.pow(hex / 255.0, GAMMA);
    }
  }

  private int toPaletteIndex(double intensity) {
    return bindInt((int)Math.round(255 * intensity), 0, 255);
  }

  private boolean checkerBoardTest(double x, double y, double size) {
    x /= size;
    y /= size;
    int X = (int)Math.floor(x);
    int Y = (int)Math.floor(y);
    return ((X + Y) & 1) == 1;
  }

  private double fractionalBrownianMotion(
      double x, double y, double z, double[] fbmMaterial) {

    double gain = 1.0;
    double lacunarity = fbmMaterial[FBM_SCALE];
    double sum = 0.0;

    for(int i = 0; i < FBM_ITERATIONS;
        i++, gain *= fbmMaterial[FBM_GAIN],
        lacunarity *= fbmMaterial[FBM_LACUNARITY]) {
      sum += gain * noiseFunction(
          x * lacunarity, y * lacunarity, z * lacunarity);
    }

    return sum;
  }

  private double noiseFunction(double x, double y, double z) {

    int floorX = (int)Math.floor(x);
    int floorY = (int)Math.floor(y);
    int floorZ = (int)Math.floor(z);

    double percentX = x - floorX;
    double percentY = y - floorY;
    double percentZ = z - floorZ;

    int x1 = floorX & NOISE_MASK;
    int y1 = floorY & NOISE_MASK;
    int z1 = floorZ & NOISE_MASK;
    int x2 = (x1 + 1) & NOISE_MASK;
    int y2 = (y1 + 1) & NOISE_MASK;
    int z2 = (z1 + 1) & NOISE_MASK;

    double n111 = noise[z1][y1][x1];
    double n211 = noise[z1][y1][x2];
    double n121 = noise[z1][y2][x1];
    double n221 = noise[z1][y2][x2];

    double n112 = noise[z2][y1][x1];
    double n212 = noise[z2][y1][x2];
    double n122 = noise[z2][y2][x1];
    double n222 = noise[z2][y2][x2];

    double n11 = n111 + (n112 - n111) * percentZ;
    double n21 = n211 + (n212 - n211) * percentZ;
    double n12 = n121 + (n122 - n121) * percentZ;
    double n22 = n221 + (n222 - n221) * percentZ;

    double a = n11 + (n12 - n11) * percentY;
    double b = n21 + (n22 - n21) * percentY;

    return a + (b - a) * percentX;
  }

  private void initializeNoise() {
    Random random = new Random(NOISE_SEED);
    for(int i = 0; i < NOISE_SIZE; i++) {
      for(int j = 0; j < NOISE_SIZE; j++) {
        for(int k = 0; k < NOISE_SIZE; k++) {
          noise[i][j][k] = random.nextDouble();
        }
      }
    }
  }

  private void saveImage() throws Throwable {
    int[] data = new int[WIDTH * HEIGHT];

    for(int y = 0, k = 0; y < HEIGHT; y++) {
      for(int x = 0; x < WIDTH; x++, k++) {
        int value = 0;
        for(int i = 0; i < 3; i++) {
          int intensity = (int)Math.round(255
              * Math.pow(pixels[y][x][i] * INVERSE_SAMPLES, INVERSE_GAMMA));
          if (intensity < 0) {
            intensity = 0;
          } else if (intensity > 255) {
            intensity = 255;
          }
          value <<= 8;
          value |= intensity;
        }
        data[k] = value;
      }
    }

    BufferedImage image = new BufferedImage(
        WIDTH, HEIGHT, BufferedImage.TYPE_INT_RGB);
    image.setRGB(0, 0, WIDTH, HEIGHT, data, 0, WIDTH);
    ImageIO.write(image, IMAGE_TYPE, new File(OUTPUT_FILE));
  }

  public static void main(String... args) throws Throwable {

    long startTime = System.currentTimeMillis();

    Main main = new Main();
    main.launch();

    long interval = System.currentTimeMillis() - startTime;
    long hours = interval / HOUR_MILLIS;
    interval %= HOUR_MILLIS;
    long minutes = interval / MINUTE_MILLIS;
    interval %= MINUTE_MILLIS;
    long seconds = interval / SECOND_MILLIS;
    interval %= SECOND_MILLIS;
    System.out.format("%d hour%s, %d minute%s, %d second%s, %d millisecond%s%n",
        hours, hours == 1 ? "" : "s",
        minutes, minutes == 1 ? "" : "s",
        seconds, seconds == 1 ? "" : "s",
        interval, interval == 1 ? "" : "s");
  }

}
